# -*- coding: utf-8 -*-
import os
from contextlib import contextmanager

import mlflow
from mlflow.exceptions import MlflowException

from doc_ufcn.train import logger


def setup_environment(config):
    needed_variables = {
        "MLFLOW_S3_ENDPOINT_URL": "s3_endpoint_url",
        "MLFLOW_TRACKING_URI": "tracking_uri",
        "AWS_ACCESS_KEY_ID": "aws_access_key_id",
        "AWS_SECRET_ACCESS_KEY": "aws_secret_access_key",
    }
    for variable_name, config_key in needed_variables.items():
        if config_key in config:
            os.environ[variable_name] = config[config_key]


@contextmanager
def start_mlflow_run(config):
    # Set needed variables in environment
    setup_environment(config)

    # Set experiment from config
    experiment_id = config.get("experiment_id")
    assert experiment_id, "Missing MLflow experiment ID in the configuration"

    try:
        mlflow.set_experiment(experiment_id=experiment_id)
    except MlflowException as e:
        logger.error(f"Couldn't set Mlflow experiment with ID: {experiment_id}")
        raise e

    # Start run
    yield mlflow.start_run(run_name=config.get("run_name"))



# -*- coding: utf-8 -*-

"""
    The run experiment module
    ======================

    Use it to train, predict and evaluate a model.
"""

import logging
import os
from pathlib import Path

from torch.cuda.amp import GradScaler
from torch.optim import Adam
from torch.utils.data import DataLoader
from torchvision import transforms
from tqdm import tqdm

from doc_ufcn import model
from doc_ufcn.models import download_model
from doc_ufcn.train.evaluate import run as evaluate
from doc_ufcn.train.mlflow_utils import start_mlflow_run
from doc_ufcn.train.normalization_params import run as normalization_params
from doc_ufcn.train.predict import run as predict
from doc_ufcn.train.training import run as train
from doc_ufcn.train.utils import DLACollateFunction, Sampler
from doc_ufcn.train.utils.preprocessing import (
    Normalize,
    Pad,
    PredictionDataset,
    Rescale,
    ToTensor,
    TrainingDataset,
)
from doc_ufcn.train.utils.training import Diceloss
from doc_ufcn.utils import export_list

logger = logging.getLogger(__name__)


def get_mean_std(log_path: Path, mean_name: str, std_name: str) -> dict:
    """
    Retrieve the mean and std values computed during the first 'normalization
    params' step.
    :param log_path: Path to save the experiment information and model.
    :param mean_name: Name of the file that will contain all the mean values.
    :param std_name: Name of the file that will contain all the std values.
    :return: A dictionary containing the mean and std values.
    """
    mean_path = log_path / mean_name
    if not mean_path.exists():
        raise Exception(f"No file found at {mean_path}")

    std_path = log_path / std_name
    if not std_path.exists():
        raise Exception(f"No file found at {std_path}")

    with mean_path.open() as f:
        mean = f.read().splitlines()
        mean = [int(value) for value in mean]

    with std_path.open() as f:
        std = f.read().splitlines()
        std = [int(value) for value in std]

    return {"mean": mean, "std": std}


def training_loaders(
    norm_params: dict,
    exp_data_paths: dict,
    classes_colors: list,
    img_size: int,
    bin_size: int,
    batch_size: int,
    no_of_params: int,
    num_workers: int = 2,
) -> dict:
    """
    Generate the loaders to use during the training step.
    :param norm_params: The mean and std values used during image normalization.
    :param exp_data_paths: Path to the data folders.
    :return loaders: A dictionary with the loaders.
    """
    loaders = {}
    t = tqdm(["train", "val"])
    t.set_description("Loading data")
    for set, images, masks in zip(
        t,
        [exp_data_paths["train"]["image"], exp_data_paths["val"]["image"]],
        [exp_data_paths["train"]["mask"], exp_data_paths["val"]["mask"]],
    ):
        dataset = TrainingDataset(
            images,
            masks,
            classes_colors,
            transform=transforms.Compose(
                [
                    Rescale(img_size),
                    Normalize(norm_params["mean"], norm_params["std"]),
                ]
            ),
        )
        loaders[set] = DataLoader(
            dataset,
            num_workers=num_workers,
            pin_memory=True,
            batch_sampler=Sampler(
                dataset,
                bin_size=bin_size,
                batch_size=batch_size,
                nb_params=no_of_params,
            ),
            collate_fn=DLACollateFunction(),
        )
        logging.info(f"{set}: Found {len(dataset)} images")
    return loaders


def prediction_loaders(
    norm_params: dict, exp_data_paths: dict, img_size: int, num_workers: int = 2
) -> dict:
    """
    Generate the loaders to use during the prediction step.
    :param norm_params: The mean and std values used during image normalization.
    :param exp_data_paths: Path to the data folders.
    :return loaders: A dictionary with the loaders.
    """
    loaders = {}
    for set, images in zip(
        ["train", "val", "test"],
        [
            exp_data_paths["train"]["image"],
            exp_data_paths["val"]["image"],
            exp_data_paths["test"]["image"],
        ],
    ):
        dataset = PredictionDataset(
            images,
            transform=transforms.Compose(
                [
                    Rescale(img_size),
                    Normalize(norm_params["mean"], norm_params["std"]),
                    Pad(),
                    ToTensor(),
                ]
            ),
        )
        loaders[set + "_loader"] = DataLoader(
            dataset,
            batch_size=1,
            shuffle=False,
            num_workers=num_workers,
            pin_memory=True,
        )
    return loaders


def training_initialization(
    training: dict,
    log_path: str,
    classes_names: list,
    use_amp: bool,
    learning_rate: float,
) -> dict:
    """
    Initialize the training step.
    :param training: Training parameters.
    :param log_path: Path to save the experiment information and model.
    :return tr_params: A dictionary with the training parameters.
    """
    no_of_classes = len(classes_names)
    # TODO: log number of classes on tensorboard ?
    net = model.load_network(no_of_classes, use_amp)
    net.apply(model.weights_init)

    if training["restore_model"] is None:
        tr_params = {
            "net": net,
            "criterion": Diceloss(no_of_classes),
            "optimizer": Adam(net.parameters(), lr=learning_rate),
            "saved_epoch": 0,
            "best_loss": 10e5,
            "scaler": GradScaler(enabled=use_amp),
            "use_amp": use_amp,
        }
    else:
        # Restore model to resume training.
        checkpoint, net, optimizer, scaler = model.restore_model(
            net,
            Adam(net.parameters(), lr=learning_rate),
            GradScaler(enabled=use_amp),
            log_path,
            training["restore_model"],
            training["same_classes"],
        )
        tr_params = {
            "net": net,
            "criterion": Diceloss(no_of_classes),
            "optimizer": optimizer,
            "saved_epoch": checkpoint["epoch"],
            "best_loss": checkpoint["best_loss"]
            if training["loss"] == "best"
            else 10e5,
            "scaler": scaler,
            "use_amp": use_amp,
        }
    return tr_params


def prediction_initialization(
    model_path: Path, classes_names: list, log_path: str
) -> dict:
    """
    Initialize the prediction step.
    :param model_path: Path of the model to load in memory.
    :param log_path: Path to save the experiment information and model.
    :return: A dictionary with the prediction parameters.
    """
    no_of_classes = len(classes_names)
    net = model.load_network(no_of_classes, False)

    _, net, _, _ = model.restore_model(net, None, None, log_path, model_path)
    return net


def run(config: dict, num_workers: int = 2):
    """
    Main program, training a new model, using a valid configuration
    """
    assert len(config["steps"]) > 0, "No step to run"
    if config["mlflow"]:
        with start_mlflow_run(config["mlflow"]) as run:
            logger.info(f"Started MLflow run with ID ({run.info.run_id})")
            run_experiment(config=config, num_workers=num_workers, mlflow_logging=True)
    else:
        run_experiment(config=config, num_workers=num_workers, mlflow_logging=False)


def run_experiment(config: dict, num_workers: int = 2, mlflow_logging=False):
    """
    Main program, training a new model, using a valid configuration
    """
    assert len(config["steps"]) > 0, "No step to run"

    if "normalization_params" in config["steps"]:
        normalization_params(
            log_path=config["log_path"],
            data_paths=config["data_paths"],
            img_size=config["img_size"],
            mean_name=config["mean"],
            std_name=config["std"],
            num_workers=num_workers,
        )

    model_to_restore = config["training"].get("restore_model", "")

    if model_to_restore and not model_to_restore.endswith(".pth"):
        # Try to load a model with such a name on HuggingFace
        logger.info(f"Loading model with name {model_to_restore} from HuggingFace")
        # Store the path to the last model and its parameters
        config["training"]["restore_model"], parameters = download_model(
            name=model_to_restore
        )
        # Store mean and std values
        export_list(data=parameters["mean"], output=config["log_path"] / config["mean"])
        export_list(data=parameters["std"], output=config["log_path"] / config["std"])

    if "train" in config["steps"] or "prediction" in config["steps"]:
        # Get the mean and std values.
        norm_params = get_mean_std(config["log_path"], config["mean"], config["std"])

    if "train" in config["steps"]:
        # Generate the loaders and start training.
        loaders = training_loaders(
            norm_params=norm_params,
            exp_data_paths=config["data_paths"],
            classes_colors=config["classes_colors"],
            img_size=config["img_size"],
            bin_size=config["bin_size"],
            batch_size=config["batch_size"],
            no_of_params=config["no_of_params"],
            num_workers=num_workers,
        )

        tr_params = training_initialization(
            config["training"],
            config["log_path"],
            config["classes_names"],
            config["use_amp"],
            config["learning_rate"],
        )
        train(
            config["model_path"],
            config["log_path"],
            config["tb_path"],
            config["no_of_epochs"],
            norm_params,
            config["classes_names"],
            loaders,
            tr_params,
            mlflow_logging,
        )

    if "prediction" in config["steps"]:
        # Generate the loaders and start predicting.
        loaders = prediction_loaders(
            norm_params, config["data_paths"], config["img_size"]
        )
        net = prediction_initialization(
            config["model_path"], config["classes_names"], config["log_path"]
        )
        predict(
            config["prediction_path"],
            config["log_path"],
            config["img_size"],
            config["classes_colors"],
            config["classes_names"],
            config["save_image"],
            config["min_cc"],
            loaders,
            net,
        )

    if "evaluation" in config["steps"]:
        for set in config["data_paths"].keys():
            for dataset in config["data_paths"][set]["json"]:
                if os.path.isdir(dataset):
                    evaluate(
                        config["log_path"],
                        config["classes_names"],
                        set,
                        config["data_paths"][set]["json"],
                        str(dataset.parent.parent.name),
                        config["prediction_path"],
                        config["evaluation_path"],
                        mlflow_logging,
                    )
                else:
                    logging.info(f"{dataset} folder not found.")



# -*- coding: utf-8 -*-

"""
    The utils module
    ======================

    Generic functions used during all the steps.
"""

import copy
import random

import numpy as np
import torch

# Useful functions.


def rgb_to_gray_value(rgb: tuple) -> int:
    """
    Compute the gray value of a RGB tuple.
    :param rgb: The RGB value to transform.
    :return: The corresponding gray value.
    """
    try:
        return int(rgb[0] * 0.299 + rgb[1] * 0.587 + rgb[2] * 0.114)
    except TypeError:
        return int(int(rgb[0]) * 0.299 + int(rgb[1]) * 0.587 + int(rgb[2]) * 0.114)


def rgb_to_gray_array(rgb: np.ndarray) -> np.ndarray:
    """
    Compute the gray array (NxM) of a RGB array (NxMx3).
    :param rgb: The RGB array to transform.
    :return: The corresponding gray array.
    """
    gray_array = rgb[:, :, 0] * 0.299 + rgb[:, :, 1] * 0.587 + rgb[:, :, 2] * 0.114
    return np.uint8(gray_array)


def create_buckets(images_sizes, bin_size):
    """
    Group images into same size buckets.
    :param images_sizes: The sizes of the images.
    :param bin_size: The step between two buckets.
    :return bucket: The images indices grouped by size.
    """
    max_size = max([image_size for image_size in images_sizes.values()])
    min_size = min([image_size for image_size in images_sizes.values()])

    bucket = {}
    current = min_size + bin_size - 1
    while current < max_size:
        bucket[current] = []
        current += bin_size
    bucket[max_size] = []

    for index, value in images_sizes.items():
        dict_index = (((value - min_size) // bin_size) + 1) * bin_size + min_size - 1
        bucket[min(dict_index, max_size)].append(index)

    bucket = {
        dict_index: values for dict_index, values in bucket.items() if len(values) > 0
    }
    return bucket


class Sampler(torch.utils.data.Sampler):
    def __init__(self, data, bin_size=20, batch_size=None, nb_params=None):
        self.bin_size = bin_size
        self.batch_size = batch_size
        self.nb_params = nb_params

        self.data_sizes = [image["size"] for image in data]

        self.vertical = {
            index: image["size"][1]
            for index, image in enumerate(data)
            if image["size"][0] > image["size"][1]
        }
        self.horizontal = {
            index: image["size"][0]
            for index, image in enumerate(data)
            if image["size"][0] <= image["size"][1]
        }

        self.buckets = [
            create_buckets(self.vertical, self.bin_size)
            if len(self.vertical) > 0
            else {},
            create_buckets(self.horizontal, self.bin_size)
            if len(self.horizontal) > 0
            else {},
        ]

    def __len__(self):
        return len(self.vertical) + len(self.horizontal)

    def __iter__(self):
        buckets = copy.deepcopy(self.buckets)
        for index, bucket in enumerate(buckets):
            for key in bucket.keys():
                random.shuffle(buckets[index][key])

        if self.batch_size is not None and self.nb_params is None:
            final_indices = []
            index_current = -1
            for bucket in buckets:
                current_batch_size = self.batch_size
                for key in sorted(bucket.keys(), reverse=True):
                    for index in bucket[key]:
                        if current_batch_size + 1 > self.batch_size:
                            current_batch_size = 0
                            final_indices.append([])
                            index_current += 1
                        current_batch_size += 1
                        final_indices[index_current].append(index)
            random.shuffle(final_indices)

        elif self.nb_params is not None:
            final_indices = []
            index_current = -1
            for bucket in buckets:
                current_params = self.nb_params
                for key in sorted(bucket.keys(), reverse=True):
                    for index in bucket[key]:
                        element_params = (
                            self.data_sizes[index][0] * self.data_sizes[index][1] * 3
                        )
                        if current_params + element_params > self.nb_params:
                            current_params = 0
                            final_indices.append([])
                            index_current += 1
                        current_params += element_params
                        final_indices[index_current].append(index)
            random.shuffle(final_indices)

        return iter(final_indices)


def pad_images_masks(
    images: list, masks: list, image_padding_value: int, mask_padding_value: int
):
    """
    Pad images and masks to create batchs.
    :param images: The batch images to pad.
    :param masks: The batch masks to pad.
    :param image_padding_value: The value used to pad the images.
    :param mask_padding_value: The value used to pad the masks.
    :return padded_images: An array containing the batch padded images.
    :return padded_masks: An array containing the batch padded masks.
    """
    heights = [element.shape[0] for element in images]
    widths = [element.shape[1] for element in images]
    max_height = max(heights)
    max_width = max(widths)

    # Make the tensor shape be divisible by 8.
    if max_height % 8 != 0:
        max_height = int(8 * np.ceil(max_height / 8))
    if max_width % 8 != 0:
        max_width = int(8 * np.ceil(max_width / 8))

    padded_images = (
        np.ones((len(images), max_height, max_width, images[0].shape[2]))
        * image_padding_value
    )
    padded_masks = np.ones((len(masks), max_height, max_width)) * mask_padding_value
    for index, (image, mask) in enumerate(zip(images, masks)):
        delta_h = max_height - image.shape[0]
        delta_w = max_width - image.shape[1]
        top, bottom = delta_h // 2, delta_h - (delta_h // 2)
        left, right = delta_w // 2, delta_w - (delta_w // 2)
        padded_images[
            index,
            top : padded_images.shape[1] - bottom,
            left : padded_images.shape[2] - right,
            :,
        ] = image
        padded_masks[
            index,
            top : padded_masks.shape[1] - bottom,
            left : padded_masks.shape[2] - right,
        ] = mask

    return padded_images, padded_masks


class DLACollateFunction:
    def __init__(self):
        self.image_padding_token = 0
        self.mask_padding_token = 0

    def __call__(self, batch):
        image = [item["image"] for item in batch]
        mask = [item["mask"] for item in batch]
        pad_image, pad_mask = pad_images_masks(
            image, mask, self.image_padding_token, self.mask_padding_token
        )
        return {
            "image": torch.tensor(pad_image).permute(0, 3, 1, 2),
            "mask": torch.tensor(pad_mask),
        }



# -*- coding: utf-8 -*-
import json
import os
from pathlib import Path

import cv2
import pytest
import yaml

FIXTURES = Path(__file__).resolve().parent / "data"


@pytest.fixture
def test_image():
    image = cv2.imread(os.path.join(FIXTURES, "test_image.png"))
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    return image


@pytest.fixture
def test_masked_image():
    image = cv2.imread(os.path.join(FIXTURES, "masked_image.png"))
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    return image


@pytest.fixture
def mask_image():
    image = cv2.imread(os.path.join(FIXTURES, "mask_image.png"))
    image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    return image


@pytest.fixture
def test_parameters():
    with open(os.path.join(FIXTURES, "test_params.yaml"), "r") as f:
        parameters = yaml.safe_load(f)
    return parameters


@pytest.fixture
def overlap_image():
    image = cv2.imread(os.path.join(FIXTURES, "overlap_image.png"))
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    return image


@pytest.fixture
def test_json():
    with open(os.path.join(FIXTURES, "test_json.json"), "r") as json_file:
        return json.load(json_file)


@pytest.fixture
def page_generic_model_path():
    return os.path.join(FIXTURES, "page_generic.pth")


@pytest.fixture
def expected_mask_path():
    return os.path.join(FIXTURES, "expected_mask.png")



# -*- coding: utf-8 -*-

"""
    The pixel metrics module
    ======================

    Use it to compute different metrics during evaluation.
    Available metrics:
        - Intersection-over-Union
        - Precision
        - Recall
        - F-score
"""

import numpy as np


def compute_metrics(
    labels: list, predictions: list, classes: list, global_metrics: dict
) -> dict:
    """
    Compute the pixel level metrics between prediction and label areas of
    a given page.
    :param labels: The label's polygons.
    :param predictions: The predicted polygons.
    :param classes: The classes names involved in the experiment.
    :param global_metrics: The initialized results dictionary.
    :return global_metrics: The updated results dictionary.
    """
    for channel in classes:
        inter = 0
        for _, gt in labels[channel]:
            for _, pred in predictions[channel]:
                inter += gt.intersection(pred).area
        gt_area = np.sum([gt.area for _, gt in labels[channel]])
        pred_area = np.sum([pred.area for _, pred in predictions[channel]])

        global_metrics[channel]["iou"].append(get_iou(inter, gt_area, pred_area))
        precision = get_precision(inter, pred_area)
        recall = get_recall(inter, gt_area)
        global_metrics[channel]["precision"].append(precision)
        global_metrics[channel]["recall"].append(recall)
        if precision + recall != 0:
            global_metrics[channel]["fscore"].append(
                2 * precision * recall / (precision + recall)
            )
        else:
            global_metrics[channel]["fscore"].append(0)
    return global_metrics


def get_iou(intersection: float, label_area: float, predicted_area: float) -> float:
    """
    Get the Intersection-over-Union value between prediction and label areas of
    a given page.
    :param intersection: Area of the intersection.
    :param label_area: Area of the label objects.
    :param predicted_area: Area of the predicted objects.
    :return: The computed Intersection-over-Union value.
    """
    union = label_area + predicted_area - intersection
    # Nothing to detect and nothing predicted.
    if label_area == 0 and predicted_area == 0:
        return 1
    # Objects to detect and/or predicted but no intersection.
    if intersection == 0 and union != 0:
        return 0
    # Objects to detect and/or predicted that intersect.
    return intersection / union


def get_precision(intersection: float, predicted_area: float) -> float:
    """
    Get the precision between prediction and label areas of
    a given page.
    :param intersection: Area of the intersection.
    :param predicted_area: Area of the predicted objects.
    :return: The computed precision value.
    """
    # Nothing predicted.
    if predicted_area == 0:
        return 1
    return intersection / predicted_area


def get_recall(intersection, label_area) -> float:
    """
    Get the recall between prediction and label areas of
    a given page.
    :param intersection: Area of the intersection.
    :param label_area: Area of the label objects.
    :return: The computed recall value.
    """
    # Nothing to detect.
    if label_area == 0:
        return 1
    return intersection / label_area



# -*- coding: utf-8 -*-

"""
    The evaluation utils module
    ======================

    Use it to during the evaluation stage.
"""

import json
import os

import matplotlib.font_manager as fm
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import ticker
from shapely.geometry import Polygon


def resize_polygons(polygons: dict, gt_size: tuple, pred_size: tuple) -> dict:
    """
    Resize the detected polygons to the original image size.
    :param polygons: The polygons to resize.
    :param gt_size: The ground truth image size.
    :param pred_size: The prediction image size.
    :return polygons: The resized detected polygons.
    """
    # Compute resizing ratio
    ratio = [gt / pred for gt, pred in zip(gt_size, pred_size)]

    for channel in polygons.keys():
        if channel == "img_size":
            continue
        for index, polygon in enumerate(polygons[channel]):
            x_points = [int(element[1] * ratio[0]) for element in polygon["polygon"]]
            y_points = [int(element[0] * ratio[1]) for element in polygon["polygon"]]

            x_points = [
                int(element) if element < gt_size[0] else int(gt_size[0])
                for element in x_points
            ]
            y_points = [
                int(element) if element < gt_size[1] else int(gt_size[1])
                for element in y_points
            ]
            x_points = [int(element) if element > 0 else 0 for element in x_points]
            y_points = [int(element) if element > 0 else 0 for element in y_points]

            assert max(x_points) <= gt_size[0]
            assert min(x_points) >= 0
            assert max(y_points) <= gt_size[1]
            assert min(y_points) >= 0
            polygons[channel][index]["polygon"] = list(zip(y_points, x_points))
    return polygons


def read_json(filename: str) -> dict:
    """
    Read a label / prediction json file.
    :param filename: Path to the file to read.
    :return: A dictionary with the file content.
    """
    with open(filename, "r") as file:
        return json.load(file)


def get_polygons(regions: dict, classes: list) -> dict:
    """
    Retrieve the polygons from a read json file.
    :param regions: Regions extracted from a label / prediction file.
    :param classes: The classes names involved in the experiment.
    :return polys: A dictionary with retrieved polygons and the
                   corresponding confidence scores.
    """
    polys = {}
    for index, channel in enumerate(classes[1:], 1):
        if channel in regions.keys():
            polys[channel] = [
                (polygon["confidence"], Polygon(polygon["polygon"]).buffer(0))
                for polygon in regions[channel]
            ]
    return polys


# Save the metrics.


def save_results(
    pixel_results: dict, object_results: dict, classes: list, path: str, dataset: str
):
    """
    Save the pixel and object results into a json file.
    :param pixel_results: The results obtained at pixel level.
    :param object_results: The results obtained at object level.
    :param classes: The classes names involved in the experiment.
    :param path: The path to the results directory.
    :param dataset: The name of the current dataset.
    """
    json_dict = {channel: {} for channel in classes}

    for channel in classes:
        json_dict[channel]["iou"] = np.round(np.mean(pixel_results[channel]["iou"]), 4)
        json_dict[channel]["precision"] = np.round(
            np.mean(pixel_results[channel]["precision"]), 4
        )
        json_dict[channel]["recall"] = np.round(
            np.mean(pixel_results[channel]["recall"]), 4
        )
        json_dict[channel]["fscore"] = np.round(
            np.mean(pixel_results[channel]["fscore"]), 4
        )
        aps = object_results[channel]["AP"]
        json_dict[channel]["AP@[.5]"] = np.round(aps[50], 4)
        json_dict[channel]["AP@[.75]"] = np.round(aps[75], 4)
        json_dict[channel]["AP@[.95]"] = np.round(aps[95], 4)
        json_dict[channel]["AP@[.5,.95]"] = np.round(np.mean(list(aps.values())), 4)

    with open(os.path.join(path, dataset + "_results.json"), "w") as json_file:
        json.dump(json_dict, json_file, indent=4)


def save_graphical_results(results: dict, classes: list, path: str):
    """
    Plot various curves involving the computed metrics.
    :param results: The dictionary containing the computed object metrics.
    :param classes: The classes names involved in the experiment.
    :param path: The path to the results directory.
    """
    plot_precision_recall_curve(results, classes, path)
    plot_rank_score(results, classes, "Precision", path)
    plot_rank_score(results, classes, "Recall", path)
    plot_rank_score(results, classes, "F-score", path)


def generate_figure(params: dict, rotation: bool = None):
    """
    Generic function to generate a figure.
    :param params: A dictionary containing the useful information
                   the initialize a figure.
    :param rotation: A boolean indicating whether to rotate the y-axis labels.
    :return fig: The initialized figure.
    :return axis: The created axis to plot.
    :return fp_light: The loaded font property used in the figures.
    """
    fp_light = fm.FontProperties(fname="./resources/font/Quicksand-Light.ttf", size=11)
    fp_medium = fm.FontProperties(
        fname="./resources/font/Quicksand-Medium.ttf", size=11
    )
    fig = plt.figure(figsize=params["size"])
    axis = fig.add_subplot(111)
    axis.set_xlabel(params["xlabel"], fontproperties=fp_light)
    axis.set_ylabel(params["ylabel"], fontproperties=fp_light)
    axis.set_xticklabels(params["xticks"], fontproperties=fp_light)
    axis.set_yticklabels(
        params["yticks"], fontproperties=fp_light, rotation=rotation, va="center"
    )
    plt.title(params["title"], fontproperties=fp_medium, fontsize=16, pad=20)
    return fig, axis, fp_light


def plot_rank_score(scores: dict, classes: list, metric: str, path: str):
    """
    Plot the scores according to the ranks on confidence score and for
    different IoU thresholds.
    :param scores: The computed scores to plot.
    :param classes: The involved classes names.
    :param metric: The current metric to plot.
    :param path: The path to the results directory.
    """
    params = {
        "size": (12, 8),
        "title": metric + " vs. confidence score for various IoU thresholds",
        "xlabel": "Confidence score",
        "ylabel": metric,
        "xticks": [0, 0.50, 0.55, 0.60, 0.65, 0.70, 0.75, 0.80, 0.85, 0.90, 0.95],
        "yticks": [0, 0.2, 0.4, 0.6, 0.8, 1],
    }
    colors = plt.cm.RdPu(np.linspace(0.2, 1, 10))
    for channel in classes:
        _, axis, fp_light = generate_figure(params)
        axis.grid(color="grey", alpha=0.2)
        axis.xaxis.set_major_locator(ticker.MultipleLocator(5))
        for index, iou in enumerate(range(50, 100, 5)):
            if metric == "Precision":
                score = list(scores[channel]["precision"][iou].values())
                rank = list(scores[channel]["precision"][iou].keys())
            if metric == "Recall":
                score = list(scores[channel]["recall"][iou].values())
                rank = list(scores[channel]["recall"][iou].keys())
            if metric == "F-score":
                score = list(scores[channel]["fscore"][iou].values())
                rank = list(scores[channel]["fscore"][iou].keys())
            axis.plot(
                rank,
                score,
                label="{:.2f}".format(iou / 100),
                alpha=1,
                color=colors[index],
                linewidth=2,
            )
            axis.scatter(
                rank,
                score,
                color=colors[index],
                facecolors="none",
                linewidth=1,
                marker="o",
            )
        axis.set_xlim([49, 96])
        axis.set_ylim([0, 1])
        plt.legend(prop=fp_light, loc="lower left")
        plt.savefig(
            os.path.join(path, metric + "_" + channel + ".png"), bbox_inches="tight"
        )


def plot_precision_recall_curve(object_metrics: dict, classes: list, path: str):
    """
    Plot the precision-recall curve for different IoU thresholds.
    :param object_metrics: The computed precisions and recalls to plot.
    :param classes: The involved classes names.
    :param path: The path to the results directory.
    """
    params = {
        "size": (12, 8),
        "title": "Precision-recall curve for various IoU thresholds",
        "xlabel": "Recall",
        "ylabel": "Precision",
        "xticks": [0, 0.2, 0.4, 0.6, 0.8, 1],
        "yticks": [0, 0.2, 0.4, 0.6, 0.8, 1],
    }
    colors = plt.cm.RdPu(np.linspace(0.2, 1, 10))
    for channel in classes:
        _, axis, fp_light = generate_figure(params)
        axis.grid(color="grey", alpha=0.2)
        for index, iou in enumerate(range(50, 100, 5)):
            current_pr = list(object_metrics[channel]["precision"][iou].values())
            current_rec = list(object_metrics[channel]["recall"][iou].values())
            axis.plot(
                current_rec,
                current_pr,
                label="{:.2f}".format(iou / 100),
                alpha=1,
                color=colors[index],
                linewidth=2,
            )
            axis.scatter(
                current_rec,
                current_pr,
                color=colors[index],
                facecolors="none",
                linewidth=1,
                marker="o",
            )
        axis.set_xlim([0, 1])
        axis.set_ylim([0, 1])
        plt.legend(prop=fp_light, loc="lower right")
        plt.savefig(
            os.path.join(path, "Precision-recall_" + channel + ".png"),
            bbox_inches="tight",
        )



# -*- coding: utf-8 -*-

"""
    The train module
    ======================

    Use it to train a model.
"""

import logging
import os
import time

import mlflow
import numpy as np
import torch
from torch.cuda.amp import autocast
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm

import doc_ufcn.train.utils.training as tr_utils
import doc_ufcn.train.utils.training_pixel_metrics as p_metrics
from doc_ufcn import model


def init_metrics(no_of_classes: int) -> dict:
    """
    Initialize the epoch metrics.
    :param no_of_classes: The number of classes involved in the experiment.
    :return: A dictionary containing the initialized metrics.
    """
    return {"matrix": np.zeros((no_of_classes, no_of_classes)), "loss": 0}


def log_metrics(epoch: int, metrics: dict, writer, step: str, mlflow_logging: bool):
    """
    Log the computed metrics to Tensorboard and Omniboard.
    :param epoch: The current epoch.
    :param metrics: The metrics to log.
    :param writer: The Tensorboard object to log information.
    :param step: String indicating whether to log training or validation metrics.
    :param mlflow_logging: Whether we should log data to MLflow.
    """
    prefixed_metrics = {step + "_" + key: value for key, value in metrics.items()}

    step_name = "TRAIN" if step == "Training" else "VALID"
    for tag, scalar in prefixed_metrics.items():
        writer.add_scalar(tag, scalar, epoch)
        logging.info("  {} {}: {}={}".format(step_name, epoch, tag, round(scalar, 4)))

    if mlflow_logging:
        mlflow.log_metrics(metrics=prefixed_metrics, step=epoch)


def run_one_epoch(
    loader,
    params: dict,
    writer,
    epochs: list,
    no_of_epochs: int,
    device: str,
    norm_params: dict,
    classes_names: list,
    step: str,
):
    """
    Run one epoch of training (or validation).
    :param loader: The loader containing the images and masks.
    :param params: A dictionary containing all the training parameters.
    :param writer: The Tensorboard object to log information.
    :param epochs: A list containing the current epoch number and the last saved epoch.
    :param no_of_epochs: The number of epochs to run.
    :param device: The device to run the experiment.
    :param norm_params: The mean and std values used during image normalization.
    :param classes_names: The names of the classes involved during the experiment.
    :param step: String indicating whether to run a training or validation step.
    :return params: The updated training parameters.
    :return epoch_values: The metrics computed during the epoch.
    """
    metrics = init_metrics(len(classes_names))
    epoch = epochs[0]

    t = tqdm(loader)
    step_name = "TRAIN" if step == "Training" else "VALID"
    t.set_description(
        "{} (prog) {}/{}".format(step_name, epoch, no_of_epochs + epochs[1])
    )

    for index, data in enumerate(t, 1):
        params["optimizer"].zero_grad()
        with autocast(enabled=params["use_amp"]):
            if params["use_amp"]:
                output = params["net"](data["image"].to(device).half())
            else:
                output = params["net"](data["image"].to(device).float())
            loss = params["criterion"](output, data["mask"].to(device).long())

        for pred in range(output.shape[0]):
            current_pred = np.argmax(
                output[pred, :, :, :].cpu().detach().numpy(), axis=0
            )
            current_label = data["mask"][pred, :, :].cpu().detach().numpy()
            batch_metrics = p_metrics.compute_metrics(
                current_pred, current_label, loss.item(), classes_names
            )
            metrics = p_metrics.update_metrics(metrics, batch_metrics)

        epoch_values = tr_utils.get_epoch_values(metrics, classes_names, index + 1)
        display_values = epoch_values
        display_values["loss"] = round(display_values["loss"], 4)
        t.set_postfix(values=str(display_values))

        if step == "Training":
            params["scaler"].scale(loss).backward()
            params["scaler"].step(params["optimizer"])
            params["scaler"].update()
            # Display prediction images in Tensorboard all 100 mini-batches.
            if index == 1 or index % 100 == 99:
                tr_utils.display_training(
                    output, data["image"], data["mask"], writer, epoch, norm_params
                )

    if step == "Training":
        return params, epoch_values
    else:
        return epoch_values


def run(
    model_path: str,
    log_path: str,
    tb_path: str,
    no_of_epochs: int,
    norm_params: dict,
    classes_names: list,
    loaders: dict,
    tr_params: dict,
    mlflow_logging: bool,
):
    """
    Run the training.
    :param model_path: The path to save the trained model.
    :param log_path: Path to save the experiment information and model.
    :param tb_path: Path to save the Tensorboard events.
    :param no_of_epochs: Total number of epochs to run.
    :param norm_params: The mean and std values used during image normalization.
    :param classes_names: The names of the classes involved during the experiment.
    :param loaders: The loaders containing the images and masks to use.
    :param tr_params: The training parameters.
    :param mlflow_logging: Whether we should log data to MLflow.
    """
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

    # Run training.
    writer = SummaryWriter(os.path.join(log_path, tb_path))
    logging.info("Starting training")
    starting_time = time.time()

    for epoch in range(1, no_of_epochs + 1):
        current_epoch = epoch + tr_params["saved_epoch"]
        # Run training.
        tr_params["net"].train()
        tr_params, epoch_values = run_one_epoch(
            loaders["train"],
            tr_params,
            writer,
            [current_epoch, tr_params["saved_epoch"]],
            no_of_epochs,
            device,
            norm_params,
            classes_names,
            step="Training",
        )

        log_metrics(
            epoch=current_epoch,
            metrics=epoch_values,
            writer=writer,
            step="Training",
            mlflow_logging=mlflow_logging,
        )

        with torch.no_grad():
            # Run evaluation.
            tr_params["net"].eval()
            epoch_values = run_one_epoch(
                loaders["val"],
                tr_params,
                writer,
                [current_epoch, tr_params["saved_epoch"]],
                no_of_epochs,
                device,
                norm_params,
                classes_names,
                step="Validation",
            )
            log_metrics(
                current_epoch,
                epoch_values,
                writer,
                step="Validation",
                mlflow_logging=mlflow_logging,
            )
            # Keep best model.
            if epoch_values["loss"] < tr_params["best_loss"]:
                tr_params["best_loss"] = epoch_values["loss"]
                model.save_model(
                    current_epoch + 1,
                    tr_params["net"].state_dict(),
                    epoch_values["loss"],
                    tr_params["optimizer"].state_dict(),
                    tr_params["scaler"].state_dict(),
                    (log_path / model_path).absolute(),
                )
                logging.info("Best model (epoch %d) saved", current_epoch)

    # Save last model.
    path = str(log_path / f"last_{model_path}").replace("model", "model_0")
    index = 1
    while os.path.exists(path):
        path = path.replace(str(index - 1), str(index))
        index += 1

    model.save_model(
        current_epoch,
        tr_params["net"].state_dict(),
        epoch_values["loss"],
        tr_params["optimizer"].state_dict(),
        tr_params["scaler"].state_dict(),
        path,
    )
    logging.info("Last model (epoch %d) saved", current_epoch)

    end = time.gmtime(time.time() - starting_time)
    logging.info(
        "Finished training in %2d:%2d:%2d", end.tm_hour, end.tm_min, end.tm_sec
    )



# -*- coding: utf-8 -*-

import os

import numpy as np
import pytest

from doc_ufcn import main
from tests.conftest import FIXTURES


@pytest.mark.parametrize(
    "no_of_classes",
    [
        (-1),
        ([]),
        (2.5),
        ("no_of_classes"),
    ],
)
def test_DocUFCN_wrong_no_of_classes(no_of_classes):
    """
    Test of the DocUFCN init function: check that a wrong number of classes raises an exception.
    """
    with pytest.raises(AssertionError):
        main.DocUFCN(no_of_classes, 768, "cpu")


@pytest.mark.parametrize(
    "no_of_classes",
    [
        (1),
        (5),
        (12),
    ],
)
def test_DocUFCN_correct_no_of_classes(no_of_classes):
    """
    Test of the DocUFCN init function: check that object is correctly created.
    """
    model = main.DocUFCN(no_of_classes, 768, "cpu")
    assert model.no_of_classes == no_of_classes
    assert model.model_input_size == 768
    assert model.device == "cpu"


@pytest.mark.parametrize(
    "input_size",
    [
        (-1),
        ([]),
        (2.5),
        ("input_size"),
    ],
)
def test_DocUFCN_wrong_input_size(input_size):
    """
    Test of the DocUFCN init function: check that a wrong input size raises an exception.
    """
    with pytest.raises(AssertionError):
        main.DocUFCN(2, input_size, "cpu")


@pytest.mark.parametrize(
    "input_size",
    [
        (768),
        (250),
    ],
)
def test_DocUFCN_correct_input_size(input_size):
    """
    Test of the DocUFCN init function: check that the object is correctly created.
    """
    model = main.DocUFCN(2, input_size, "cpu")
    assert model.no_of_classes == 2
    assert model.model_input_size == input_size
    assert model.device == "cpu"


@pytest.mark.parametrize(
    "model_path",
    [
        # Wrong path/file
        ("page_generic.pth"),
        (1),
        # Correct path
        (os.path.join(FIXTURES, "page_generic.pth")),
    ],
)
def test_load(model_path):
    """
    Test of the load function.
    """
    model = main.DocUFCN(2, 768, "cpu")

    if not os.path.isfile(model_path):
        with pytest.raises(AssertionError):
            model.load(model_path, [190, 182, 165], [48, 48, 45])
    elif not isinstance(model_path, str):
        with pytest.raises(AttributeError):
            model.load(model_path, [190, 182, 165], [48, 48, 45])
    else:
        model.load(model_path, [190, 182, 165], [48, 48, 45])
        assert model.net is not None
        assert model.mean == [190, 182, 165]
        assert model.std == [48, 48, 45]


@pytest.mark.parametrize(
    "mean",
    [
        (10),
        ([]),
        ([10]),
        (["test"]),
        ([10, "test"]),
        (-1.5),
        ("mean"),
    ],
)
def test_mean(mean, page_generic_model_path):
    """
    Test of the load function: check that wrong mean values raise an exception.
    """
    model = main.DocUFCN(2, 768, "cpu")
    with pytest.raises(AssertionError):
        model.load(page_generic_model_path, mean, [48, 48, 45])


@pytest.mark.parametrize(
    "std",
    [
        (10),
        ([]),
        ([10]),
        (["test"]),
        ([10, "test"]),
        (-1.5),
        ("mean"),
    ],
)
def test_std(std, page_generic_model_path):
    """
    Test of the load function: check that wrong std values raise an exception.
    """
    model = main.DocUFCN(2, 768, "cpu")
    with pytest.raises(AssertionError):
        model.load(page_generic_model_path, [190, 182, 165], std)


@pytest.mark.parametrize(
    "input_image",
    [([1]), ("input_image"), (1)],
)
def test_predict_input_image(input_image, page_generic_model_path):
    """
    Test of the predict function: check that wrong input images raise an exception.
    """
    model = main.DocUFCN(2, 768, "cpu")
    model.load(page_generic_model_path, [190, 182, 165], [48, 48, 45])

    with pytest.raises(AssertionError):
        model.predict(input_image)


def test_predict(test_image, test_json, page_generic_model_path):
    """
    Test of the predict function.
    """
    model = main.DocUFCN(2, 768, "cpu")
    model.load(page_generic_model_path, [190, 182, 165], [48, 48, 45])

    # Only one output
    output = model.predict(test_image)
    for key in output[0].keys():
        assert len(output[0][key]) == len(test_json[str(key)])
        for polygon, expected_polygon in zip(output[0][key], test_json[str(key)]):
            polygon["polygon"] = [list(element) for element in polygon["polygon"]]
            assert polygon["polygon"] == expected_polygon["polygon"]
            assert polygon["confidence"] == expected_polygon["confidence"]
    assert output[1:4] == (None, None, None)


def test_predict_all_outputs(
    test_image, mask_image, overlap_image, page_generic_model_path
):
    """
    Test of the predict function with all four outputs.
    """
    model = main.DocUFCN(2, 768, "cpu")
    model.load(page_generic_model_path, [190, 182, 165], [48, 48, 45])

    # All outputs
    output = model.predict(
        test_image, raw_output=True, mask_output=True, overlap_output=True
    )
    assert isinstance(output[0], dict)
    assert output[1].shape == (2, test_image.shape[0], test_image.shape[1])
    assert np.array_equal(output[2].astype(np.uint8), mask_image)
    assert np.array_equal(output[3], overlap_image)



# -*- coding: utf-8 -*-

import cv2
import numpy as np
from torch import from_numpy


def resize(input_image, network_size, padding):
    """
    Resize the input image into the network input size.
    Resize the image such that the longest side is equal to the network
    input size. Pad the image such that it is divisible by 8.
    :param input_image: The input image to resize.
    :param network_size: The input size of the model.
    :param padding: The value to use as padding.
    :return: The resized input image and the padding sizes.
    """
    old_size = input_image.shape[:2]
    if max(old_size) != network_size:
        # Compute the new sizes.
        ratio = float(network_size) / max(old_size)
        new_size = tuple([int(x * ratio) for x in old_size])
        # Resize the image.
        resized_image = cv2.resize(input_image, (new_size[1], new_size[0]))
    else:
        new_size = old_size
        resized_image = input_image

    delta_w = 0
    delta_h = 0
    if resized_image.shape[0] % 8 != 0:
        delta_h = int(8 * np.ceil(resized_image.shape[0] / 8)) - resized_image.shape[0]
    if resized_image.shape[1] % 8 != 0:
        delta_w = int(8 * np.ceil(resized_image.shape[1] / 8)) - resized_image.shape[1]
    top, bottom = delta_h // 2, delta_h - (delta_h // 2)
    left, right = delta_w // 2, delta_w - (delta_w // 2)
    resized_image = cv2.copyMakeBorder(
        resized_image, top, bottom, left, right, cv2.BORDER_CONSTANT, value=padding
    )
    return resized_image, [top, left]


def preprocess_image(input_image, model_input_size, mean, std):
    """
    Preprocess the input image before feeding it to the network.
    The image is first resized, normalized and converted to a tensor.
    :param input_image: The input image to preprocess.
    :param model_input_size: The size of the model input.
    :param mean: The mean value used to normalize the image.
    :param std: The standard deviation used to normalize the image.
    :return: The resized, normalized and padded input tensor.
    """
    # Resize the image
    resized_image, padding = resize(input_image, model_input_size, padding=mean)
    # Normalize the image
    normalized_image = np.zeros(resized_image.shape)
    for channel in range(resized_image.shape[2]):
        normalized_image[:, :, channel] = (
            np.float32(resized_image[:, :, channel]) - mean[channel]
        ) / std[channel]
    # To tensor
    normalized_image = normalized_image.transpose((2, 0, 1))
    normalized_image = np.expand_dims(normalized_image, axis=0)
    return from_numpy(normalized_image), padding



# -*- coding: utf-8 -*-
from itertools import combinations
from pathlib import Path

import cv2
import numpy
from shapely.affinity import scale
from shapely.geometry import LineString, Polygon
from shapely.geometry.multipolygon import MultiPolygon


def new_image_size(input_size: tuple, output_size: int):
    """
    Get the resize label size and corresponding ratios.
    The longest side of the image will have a length of output_size.
    :param input_size: The original image size.
    :param output_size: The output maximum size.
    """
    # Compute the new image size.
    ratio = float(output_size) / max(input_size[:2])
    new_size = tuple([int(x * ratio) for x in input_size[:2]])
    # Compute the ratios between the original and new image sizes.
    # It is used to resize the annotations.
    ratios = [new_size[index] / input_size[index] for index in range(len(new_size))]
    return *new_size, *ratios


def generate_mask(
    image_width: int,
    image_height: int,
    max_image_size: float,
    label_polygons: dict,
    label_colors: dict,
    output_path: Path,
) -> str:
    """
    Generate a mask with the given dimensions and polygons.
    Returns the path to the generated image.
    """
    if max_image_size is not None:
        mask_height, mask_width, ratio_height, ratio_width = new_image_size(
            [image_height, image_width],
            max_image_size,
        )
    else:
        mask_height, mask_width, ratio_height, ratio_width = (
            image_height,
            image_width,
            1,
            1,
        )
    # Remove the extension from the output_path and add the suffix
    black_img = numpy.zeros(
        (mask_height, mask_width, 3),
        dtype=numpy.uint8,
    )
    cv2.imwrite(output_path, black_img)
    img = cv2.imread(output_path)

    # Draw the polygons on the image
    for label in label_polygons:
        # Retrieve polygon color
        color = label_colors[label]
        # Retrieve corresponding polygons
        polygons = [Polygon(poly) for poly in label_polygons[label]]

        # Resize the polygons
        polygons = resize_polygons(
            polygons=polygons, height=ratio_height, width=ratio_width
        )

        # Split the polygons
        polygons = split_polygons(polygons)

        # Draw the polygons on the mask image
        draw_polygons(polygons, img, color)

    # Update the mask image on Disk
    cv2.imwrite(output_path, img)


def draw_polygons(polygons: list, img, color: tuple) -> None:
    """
    Draw the provided polygons on the image
    """
    for poly in polygons:
        if isinstance(poly, MultiPolygon):
            draw_polygons(polygons=list(poly.geoms), img=img, color=color)
        else:
            contours = [numpy.array(poly.exterior.coords).round().astype(numpy.int32)]
            for contour in contours:
                if len(contour) > 0:
                    cv2.drawContours(img, contours, 0, tuple(reversed(color)), -1)


def resize_polygons(polygons: list, height: float, width: float) -> list:
    """
    Resize the polygons.
    :param polygons: The polygons to resize.
    :param height: The ratio in the height dimension.
    :param width: The ratio in the width dimension.
    :return: A list of the resized polygons.
    """
    return [
        scale(polygon, xfact=width, yfact=height, origin=(0, 0)) for polygon in polygons
    ]


def split_polygons(polygons: list) -> list:
    """
    Split the touching and overlapping polygons.
    :param polygons: The polygons to split.
    :return polygons: The non-touching polygons.
    """
    eps = 2
    for comb in combinations(range(len(polygons)), 2):
        poly1 = polygons[comb[0]]
        poly2 = polygons[comb[1]]
        # Skip invalid polygons as they cannot be compared.
        if not poly1.is_valid or not poly2.is_valid:
            continue
        # If the two polygons intersect: first erode them, then check if they still intersect.
        if poly1.intersects(poly2):
            poly1 = poly1.buffer(-eps)
            poly2 = poly2.buffer(-eps)
            intersection = poly1.intersection(poly2)
            # If they still intersect, remove the intersection from the biggest polygon.
            if not intersection.is_empty:
                if (
                    isinstance(intersection, Polygon)
                    and intersection.area < 0.2 * poly1.area
                    and intersection.area < 0.2 * poly2.area
                    or isinstance(intersection, MultiPolygon)
                ):
                    if poly1.area > poly2.area:
                        polygons[comb[0]] = poly1.difference(intersection)
                        polygons[comb[1]] = poly2.buffer(-eps)
                    else:
                        polygons[comb[1]] = poly2.difference(intersection)
                        polygons[comb[0]] = poly1.buffer(-eps)
                elif isinstance(intersection, LineString):
                    polygons[comb[0]] = poly1.difference(intersection)
                    polygons[comb[1]] = poly2.difference(intersection)
        elif poly1.touches(poly2):
            polygons[comb[0]] = poly1.buffer(-2 * eps)
            polygons[comb[1]] = poly2.buffer(-2 * eps)
    # Erode all polygons so that they don't touch when drawn over the label image.
    polygons = [poly.buffer(-2 * eps) for poly in polygons]
    return polygons



# -*- coding: utf-8 -*-

"""
    The prediction utils module
    ======================

    Use it to during the prediction stage.
"""

import json

import cv2
import imageio as io
import numpy as np


def resize_polygons(
    polygons: dict, image_size: tuple, input_size: tuple, padding: tuple
) -> dict:
    """
    Resize the detected polygons to the original input image size.
    :param polygons: The polygons to resize.
    :param image_size: The original input image size.
    :param input_size: The network input size.
    :param padding: The padding of the input image.
    :return polygons: The resized detected polygons.
    """
    # Compute the small size image.
    ratio = float(input_size) / max(image_size)
    new_size = tuple([int(x * ratio) for x in image_size])
    # Compute resizing ratio
    ratio = [element / float(new) for element, new in zip(image_size, new_size)]

    for channel in polygons.keys():
        for index, polygon in enumerate(polygons[channel]):
            x_points = [element[0][1] for element in polygon["polygon"]]
            y_points = [element[0][0] for element in polygon["polygon"]]
            x_points = [
                int((element - padding["top"]) * ratio[0]) for element in x_points
            ]
            y_points = [
                int((element - padding["left"]) * ratio[1]) for element in y_points
            ]

            x_points = [
                int(element) if element < image_size[0] else int(image_size[0])
                for element in x_points
            ]
            y_points = [
                int(element) if element < image_size[1] else int(image_size[1])
                for element in y_points
            ]
            x_points = [int(element) if element > 0 else 0 for element in x_points]
            y_points = [int(element) if element > 0 else 0 for element in y_points]
            assert max(x_points) <= image_size[0]
            assert min(x_points) >= 0
            assert max(y_points) <= image_size[1]
            assert min(y_points) >= 0
            polygons[channel][index]["polygon"] = list(zip(y_points, x_points))
    return polygons


def compute_confidence(region: np.ndarray, probas: np.ndarray) -> float:
    """
    Compute the confidence score of a detected polygon.
    :param region: The detected polygon coordinates.
    :param probas: The probability map obtained by the network.
    :return: The confidence score of the given region.
    """
    mask = np.zeros(probas.shape)
    cv2.drawContours(mask, [region], 0, 1, -1)
    confidence = np.sum(mask * probas) / np.sum(mask)
    return round(confidence, 4)


# Save the prediction coordinates and images.


def save_prediction(polygons: dict, filename: str):
    """
    Save the detected polygon coordinates and their confidence score
    into a file.
    :param polygons: The detected polygons coordinates and
                     confidence scores.
    :param filename: The filename to save the detected polygons.
    """
    with open(filename.replace("png", "json"), "w") as outfile:
        json.dump(polygons, outfile, indent=4)


def save_prediction_image(polygons, colors, input_size, filename: str):
    """
    Save the detected polygon to an image.
    :param polygons: The detected polygons coordinates.
    :param colors: The colors corresponding to each involved class.
    :param input_size: The original input image size.
    :param filename: The filename to save the prediction image.
    """
    image = np.zeros((input_size[0], input_size[1], 3))
    index = 1
    for channel in polygons.keys():
        if channel == "img_size":
            continue
        color = [int(element) for element in colors[index]]
        for polygon in polygons[channel]:
            cv2.drawContours(image, [np.array(polygon["polygon"])], 0, color, -1)
        index += 1
    io.imsave(filename, np.uint8(image))



# -*- coding: utf-8 -*-
import tempfile

import cv2
import pytest
from skimage.metrics import structural_similarity as ssim

from doc_ufcn.train.mask import generate_mask


@pytest.mark.parametrize(
    "image_width, image_height, label_polygons, label_colors",
    [
        (
            400,
            500,
            {
                "text_line": [
                    [[41, 280], [361, 280], [361, 311], [41, 311], [41, 280]]
                ],
                "paragraph": [[[58, 76], [307, 76], [307, 205], [58, 205], [58, 76]]],
            },
            {"text_line": (255, 0, 0), "paragraph": (0, 0, 255)},
        )
    ],
)
def test_generate_mask(
    image_width, image_height, label_polygons, label_colors, expected_mask_path
):
    # Read the expected output image
    expected_mask = cv2.imread(expected_mask_path)

    # Generate the mask
    _, output_path = tempfile.mkstemp(suffix=".teklia.test.mask.png")
    generate_mask(
        image_width, image_height, None, label_polygons, label_colors, output_path
    )
    generated_mask = cv2.imread(output_path)

    assert generated_mask.shape == expected_mask.shape

    # Image comparison is done with skimage.structural_similarity
    similarity = ssim(
        expected_mask,
        generated_mask,
        data_range=generated_mask.max() - generated_mask.min(),
        channel_axis=2,
    )
    assert similarity >= 0.95



# -*- coding: utf-8 -*-

"""
    The training utils module
    ======================

    Use it to during the training stage.
"""

import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.nn as nn

import doc_ufcn.train.utils.training_pixel_metrics as p_metrics

# Useful functions.


class Diceloss(nn.Module):
    """
    The Diceloss class is used during training.
    """

    def __init__(self, num_classes: int):
        """
        Constructor of the Diceloss class.
        :param num_classes: The number of classes involved in the experiment.
        """
        super(Diceloss, self).__init__()
        self.num_classes = num_classes

    def forward(self, pred: np.ndarray, target: np.ndarray) -> float:
        """
        Compute the Dice loss between a label and a prediction mask.
        :param pred: The prediction made by the network.
        :param target: The label mask.
        :return: The Dice loss.
        """
        label = (
            nn.functional.one_hot(target, num_classes=self.num_classes)
            .permute(0, 3, 1, 2)
            .contiguous()
        )

        smooth = 1.0
        iflat = pred.contiguous().view(-1)
        tflat = label.contiguous().view(-1)
        intersection = (iflat * tflat).sum()
        A_sum = torch.sum(iflat * iflat)
        B_sum = torch.sum(tflat * tflat)
        return 1 - ((2.0 * intersection + smooth) / (A_sum + B_sum + smooth))


# Plot the prediction during training.


def plot_prediction(output: np.ndarray) -> np.ndarray:
    """
    Transform the output of the network into an array of categorical
    predictions.
    :param output: The predictions of the batch images.
    :return prediction: The array of categorical predictions.
    """
    prediction = np.zeros((output.shape[0], 1, output.shape[2], output.shape[3]))
    for pred in range(output.shape[0]):
        current_pred = output[pred, :, :, :]
        new = np.argmax(current_pred, axis=0)
        new = np.expand_dims(new, axis=0)
        prediction[pred, :, :, :] = new
    return prediction


def display_training(
    output: np.ndarray,
    image: np.ndarray,
    label: np.ndarray,
    writer,
    epoch: int,
    norm_params: list,
):
    """
    Define the figure to plot a batch images, labels and current predictions.
    Add it to Tensorboard.
    :param output: The predictions of the batch images.
    :param image: The current batch images.
    :param label: The corresponding labels.
    :param writer: The Tensorboard writer to add the figure.
    :param epoch: The current epoch.
    :param norm_params: The mean values and standard deviations used
                        to normalize the images.
    """
    predictions = plot_prediction(output.cpu().detach().numpy())
    fig, axs = plt.subplots(
        predictions.shape[0],
        3,
        figsize=(10, 3 * predictions.shape[0]),
        gridspec_kw={"hspace": 0.2, "wspace": 0.05},
    )
    for pred in range(predictions.shape[0]):
        current_input = image.cpu().numpy()[pred, :, :, :]
        current_input = current_input.transpose((1, 2, 0))
        for channel in range(current_input.shape[2]):
            current_input[:, :, channel] = (
                current_input[:, :, channel] * norm_params["std"][channel]
            ) + norm_params["mean"][channel]
        if predictions.shape[0] > 1:
            axs[pred, 0].imshow(current_input.astype(np.uint8))
            axs[pred, 1].imshow(label.cpu()[pred, :, :], cmap="gray")
            axs[pred, 2].imshow(predictions[pred, 0, :, :], cmap="gray")
        else:
            axs[0].imshow(current_input.astype(np.uint8))
            axs[1].imshow(label.cpu()[pred, :, :], cmap="gray")
            axs[2].imshow(predictions[pred, 0, :, :], cmap="gray")
    _ = [axi.set_axis_off() for axi in axs.ravel()]
    writer.add_figure("Image_Label_Prediction", fig, global_step=epoch)


# Display the metrics during training.


def get_epoch_values(metrics: dict, classes: list, batch: int) -> dict:
    """
    Get the metrics of an epoch.
    :param metrics: The metrics to get.
    :param classes: The classes names involved in the experiment.
    :param batch: The current batch.
    :return values: The computed epoch values.
    """
    values = {}
    for channel in classes[1:]:
        values["iou_" + channel] = round(
            p_metrics.iou(metrics["matrix"], classes.index(channel)), 6
        )
    values["loss"] = metrics["loss"] / batch
    return values



# -*- coding: utf-8 -*-

import logging
import os

import yaml
from huggingface_hub import hf_hub_download
from huggingface_hub.utils import RepositoryNotFoundError, RevisionNotFoundError

HUGGING_FACE_REPO_PREFIX = "Teklia/doc-ufcn-"


logger = logging.getLogger(__name__)


def download_model(name, version=None):
    # Strip the model name prefix if provided
    name = name.replace("doc-ufcn-", "")
    logger.info(f"Will look for model @ {HUGGING_FACE_REPO_PREFIX + name}")

    cache_dir = os.environ.get("XDG_CACHE_HOME", os.path.expanduser("~/.cache"))
    dir_path = os.path.join(cache_dir, "doc-ufcn", "models", name)

    try:
        # Retrieve parameters.yml
        parameters_path = hf_hub_download(
            repo_id=HUGGING_FACE_REPO_PREFIX + name,
            filename="parameters.yml",
            cache_dir=dir_path,
            revision=version,
        )
        # Retrieve model.pth
        model_path = hf_hub_download(
            repo_id=HUGGING_FACE_REPO_PREFIX + name,
            filename="model.pth",
            cache_dir=dir_path,
            revision=version,
        )
    except RepositoryNotFoundError as e:
        logger.error(
            f"Repository with name {name} was not found or you may not have access to it."
        )
        print(str(e))
        raise
    except RevisionNotFoundError as e:
        logger.error(
            f"Revision {version} was not found on the repository with name {name}."
        )
        print(str(e))
        raise

    with open(parameters_path) as f:
        parameters = yaml.safe_load(f)

    return model_path, parameters["parameters"]



# -*- coding: utf-8 -*-

import os

import pytest
from huggingface_hub.utils import RepositoryNotFoundError, RevisionNotFoundError

from doc_ufcn import models


@pytest.mark.parametrize(
    "name, version, expected_model_path, expected_parameters, correct_name",
    [
        # Correct name and version
        (
            "generic-page",
            "main",
            "~/.cache/doc-ufcn/models/generic-page/models--Teklia--doc-ufcn-generic-page/snapshots/ef5519d77baa190bfe294ad3994146dbe7e72095",
            pytest.lazy_fixture("test_parameters"),
            True,
        ),
        # Correct fullname and version
        (
            "doc-ufcn-generic-page",
            "main",
            "~/.cache/doc-ufcn/models/generic-page/models--Teklia--doc-ufcn-generic-page/snapshots/ef5519d77baa190bfe294ad3994146dbe7e72095",
            pytest.lazy_fixture("test_parameters"),
            True,
        ),
        # Correct name and incorrect version
        ("generic-page", "version", None, None, True),
        # Correct name and no version
        (
            "generic-page",
            None,
            "~/.cache/doc-ufcn/models/generic-page/models--Teklia--doc-ufcn-generic-page/snapshots/ef5519d77baa190bfe294ad3994146dbe7e72095",
            pytest.lazy_fixture("test_parameters"),
            True,
        ),
        # Incorrect name and incorrect version
        ("page_model", "version", None, None, False),
        # Incorrect name and no version
        ("page_model", None, None, None, False),
    ],
)
def test_download_model(
    name, version, expected_model_path, expected_parameters, correct_name
):
    """
    Test of the download_model function.
    Check that the correct model is loaded.
    """
    if expected_model_path is None and expected_parameters is None:
        if correct_name:
            # Bad model name
            with pytest.raises(RevisionNotFoundError):
                model_path, parameters = models.download_model(name, version)
        else:
            # Bad revision
            with pytest.raises(RepositoryNotFoundError):
                model_path, parameters = models.download_model(name, version)
    else:
        model_path, parameters = models.download_model(name, version)
        assert model_path == os.path.join(
            os.path.expanduser(expected_model_path), "model.pth"
        )
        assert parameters == expected_parameters



# -*- coding: utf-8 -*-

"""
    The object metrics module
    ======================

    Use it to compute different metrics during evaluation.
    Available metrics:
        - Precision
        - Recall
        - F-score
        - Average precision
"""

import numpy as np


def __compute_iou_from_contours(poly1, poly2) -> float:
    """
    Get the Intersection-over-Union value between two contours.
    :param poly1: The first contour used to compute the
                  Intersection-over-Union.
    :param poly2: The second contour used to compute the
                  Intersection-over-Union.
    :return: The computed Intersection-over-Union value between the two
             contours.
    """
    intersection = poly1.intersection(poly2).area
    union = poly1.area + poly2.area - intersection
    return intersection / union if union != 0 else 1


def __get_ious(labels: list, predictions: list) -> dict:
    """
    Get the Intersection-over-Union values of the predicted objects.
    :param labels: The coordinates of the ground truth objects.
    :param predictions: The coordinates of the predicted objects.
    :return ious: A dictionary with the highest Intersection-over-Union
                  value for each predicted object.
    """
    ious = {key: 0 for key in range(len(predictions))}

    for _, label in labels:
        best_iou = 0
        best_prediction = None
        for index, prediction in enumerate(predictions):
            iou = __compute_iou_from_contours(prediction[1], label)
            if iou > best_iou:
                best_iou = iou
                best_prediction = index
        if best_prediction is not None:
            if best_iou > ious[best_prediction]:
                ious[best_prediction] = best_iou
    return ious


def __rank_predicted_objects(labels: list, predictions: list) -> dict:
    """
    Rank the predited objects by decreasing confidence score and decreasing
    Intersection-over-Union values.
    :param labels: The coordinates and confidence scores of the ground truth objects.
    :param predictions: The coordinates and confidence scores of the predicted objects.
    :return scores: The ranked predicted objects by confidence and
                    Intersection-over-Union.
    """
    ious = __get_ious(labels, predictions)

    scores = {index: prediction[0] for index, prediction in enumerate(predictions)}
    tuples_score_iou = [(v, ious[k]) for k, v in scores.items()]
    scores = sorted(tuples_score_iou, key=lambda item: (-item[0], -item[1]))
    return scores


def compute_rank_scores(labels: list, predictions: list, classes: list) -> dict:
    """
    Compute the number of true positive objects and the total of the
    predicted objects. It is used later to compute the overall precision,
    recall and F-score.
    :param labels: The coordinates and confidence scores of the ground truth objects.
    :param predictions: The coordinates and confidence scores of the predicted objects.
    :param classes: The classes names involved in the experiment.
    :return scores: The scores obtained for a each rank, IoU
                    threshold and class.
    """
    scores = {channel: {iou: None for iou in range(50, 100, 5)} for channel in classes}
    for channel in classes:
        channel_scores = __rank_predicted_objects(labels[channel], predictions[channel])
        for iou in range(50, 100, 5):
            rank_scores = {rank: {"True": 0, "Total": 0} for rank in range(95, -5, -5)}
            for rank in range(95, -5, -5):
                rank_objects = list(
                    filter(lambda item: item[0] >= rank / 100, channel_scores)
                )
                rank_scores[rank]["True"] = sum(x[1] > iou / 100 for x in rank_objects)
                rank_scores[rank]["Total"] = len(rank_objects)
            scores[channel][iou] = rank_scores
    return scores


def update_rank_scores(global_scores: dict, image_scores: dict, classes: list) -> dict:
    """
    Update the global scores by adding page scores.
    :param global_scores: The scores obtained so far.
    :param image_scores: the current page scores.
    :param classes: The classes names involved in the experiment.
    :return global_scores: The updated global scores.
    """
    for channel in classes:
        for iou in range(50, 100, 5):
            for rank in range(95, -5, -5):
                global_scores[channel][iou][rank]["True"] += image_scores[channel][iou][
                    rank
                ]["True"]
                global_scores[channel][iou][rank]["Total"] += image_scores[channel][
                    iou
                ][rank]["Total"]
    return global_scores


def __init_results() -> dict:
    """
    Initialize the results dictionary by generating dictionary for
    the different rank and Intersection-over-Union thresholds.
    :return: The initialized results dictionary.
    """
    return {iou: {rank: 0 for rank in range(95, -5, -5)} for iou in range(50, 100, 5)}


def __get_average_precision(precisions: list, recalls: list) -> float:
    """
    Compute the mean average precision. Interpolate the precision-recall
    curve, then get the interpolated precisions for values.
    Compute the average precision.
    :param precisions: The computed precisions for a given channel and a
                       given confidence score.
    :param recalls: The computed recalls for a given channel and a given
                    confidence score.
    :return: The average precision for the channel and for the confidence
             score range.
    """
    rp_tuples = []
    # Interpolated precision-recall curve.
    while len(precisions) > 0:
        max_precision = np.max(precisions)
        argmax_precision = np.argmax(precisions)
        max_recall = recalls[argmax_precision]
        rp_tuples.append({"p": max_precision, "r": max_recall})
        for _ in range(argmax_precision + 1):
            precisions.pop(0)
            recalls.pop(0)
    rp_tuples[-1]["r"] = 1

    ps = [rp_tuple["p"] for rp_tuple in rp_tuples]
    rs = [rp_tuple["r"] for rp_tuple in rp_tuples]
    ps.insert(0, ps[0])
    rs.insert(0, 0)
    return np.trapz(ps, x=rs)


def get_mean_results(
    global_scores: dict, true_gt: dict, classes: list, results: dict
) -> dict:
    """
    Get the mean metrics values for all the set.
    :param global_scores: The overall computed scores.
    :param true_gt: The total number of ground truth objects by class.
    :param classes: The classes names involved in the experiment.
    :param results: The initialized results dictionary.
    :return results: The dictionary containing the mean computed values.
    """
    for channel in classes:
        precisions = __init_results()
        recalls = __init_results()
        fscores = __init_results()
        aps = {iou: 0 for iou in range(50, 100, 5)}
        for iou in range(50, 100, 5):
            for rank in range(95, -5, -5):
                true_predicted = global_scores[channel][iou][rank]["True"]
                predicted = global_scores[channel][iou][rank]["Total"]

                precisions[iou][rank] = (
                    true_predicted / predicted if predicted != 0 else 1
                )
                recalls[iou][rank] = (
                    true_predicted / true_gt[channel] if true_gt[channel] != 0 else 1
                )

                if precisions[iou][rank] + recalls[iou][rank] != 0:
                    fscores[iou][rank] = (
                        2
                        * (precisions[iou][rank] * recalls[iou][rank])
                        / (precisions[iou][rank] + recalls[iou][rank])
                    )
            aps[iou] = __get_average_precision(
                list(precisions[iou].values()), list(recalls[iou].values())
            )
            results[channel]["precision"] = precisions
            results[channel]["recall"] = recalls
            results[channel]["fscore"] = fscores
            results[channel]["AP"] = aps
    return results
